/*
 * mocktcp
 *
 * This library provides a reliable, ordered, stream-based, message 
 * transport designed to look very similar to TCP, although slightly
 * cleaner. Underneath, mocktcp uses rmessage reliable datagrams for
 * packet delivery. Rmessages uses UDP.
 * -Dave Lowell
 */

#include "rmessage.h"
#include "hash.h"
#include "mocktcp.h"

/**
 ** Globals
 **/

hash_table	*buf_table;


/*
 * Allocate new buffer for this connected socket.
 */
void add_buffer(int sock)
{
	conn_buffer	*b;

	rm_lock();

	b = (conn_buffer*) malloc(sizeof(*b));
	if (b == NULL) {
		fprintf(stderr, "Out of memory.\n");
		exit(1);
	}

	b->bytes = 0;
	b->index = 0;
	hash_insert(buf_table, (char*)sock, b);

	rm_unlock();
}

/*
 * Return the receive buffer for this socket, or NULL on failure.
 */
conn_buffer* get_buffer(int sock)
{
	return hash_find(buf_table, (char*) sock);
}


/*
 * Remove buffer for this socket from buffer table.
 */
void delete_buffer(int sock)
{
	conn_buffer *b;

	rm_lock();
	b = get_buffer(sock);
	free(b);
	hash_delete_entry(buf_table, (char*) sock);
	rm_unlock();
}

/*
 * Initialize mocktcp.
 */
void mt_init(void)
{
	rm_init();

	buf_table = (hash_table*) malloc(sizeof(hash_table));
	if (buf_table == NULL) {
		fprintf(stderr, "Out of memory.\n");
		exit(1);
	}
	hash_init(buf_table, HASH_ONE_WORD_KEYS);
}


/*
 * Return an open socket descriptor for use in later mocktcp calls. Most
 * standard unix operations will succeed on the returned socket descriptor.
 * Returns -1 on error.
 */
int mt_socket(void)
{
	return rm_socket();
}


/*
 * Bind an open socket to a specific port. Returns 0 on success and -1 on
 * failure.
 */
int mt_bind(int sock, int port)
{
	return rm_bind(sock, port);
}


/*
 * Waits for an incoming connection request on the socket provided,
 * and returns when one is ready. Returns 0 on success or -1 on failure.
 */
int mt_listen(int sock)
{
	int	res;
	char	b;

	res = rm_recv(sock, &b, 1, MSG_PEEK);
	return res;
}


/*
 * This function accepts a connection request on the provided socket
 * and returns a new socket descriptor once the connection is ready.
 * It blocks until a connection request comes in. It also returns the
 * address and port of the connecting client in the provided 'addr'
 * structure. Returns an open socket descriptor on success or -1 on
 * failure.
 */
int mt_accept(int sock, struct sockaddr_in *addr)
{
	int	res, newsock;
	char	b;
	struct sockaddr_in from;

	res = rm_recvfrom(sock, &b, 1, 0, &from);
	if (res < 0) return -1;

	newsock = rm_socket();
	if (newsock < 0) return -1;

	res = rm_bind(newsock, 0);
	if (res < 0) {
		close(newsock);
		return -1;
	}

	res = rm_connect_ip(newsock, from.sin_addr.s_addr,ntohs(from.sin_port));
	if (res < 0) {
		close(newsock);
		return -1;
	}

	res = rm_send(newsock, &b, 1, 0);
	if (res < 0) {
		close(newsock);
		return -1;
	}

	if (addr != NULL)
		bcopy(&from, addr, sizeof(from));
	add_buffer(newsock);

	return newsock;
}


/*
 * Open a connection with host 'host' on who is listening on port
 * 'port'. Returns 0 on success and -1 on failure.
 */
int mt_connect(int sock, char* host, int port)
{
	int	res;
	short	newport;
	char	buf;
	struct sockaddr_in to, from;
	int	myport, namelen;

	if (get_buffer(sock) != NULL) {
		errno = EISCONN;
		return -1;
	}

	res = rm_fill_sockaddr(&to, host, port);
	if (res < 0) return -1;

	res = rm_sendto(sock, "x", 1, 0, &to);
	if (res < 0) return -1;

	namelen = sizeof(to);
	res = getsockname(sock, (struct sockaddr*) &to, &namelen);
	if (res < 0) {
		perror("getsockname");
		exit(1);
	}

	res = rm_recvfrom(sock, &buf, 1, 0, &from);
	if (res < 0) return -1;

	newport = ntohs(from.sin_port);
	res = rm_connect_ip(sock, from.sin_addr.s_addr, newport);
	if (res < 0) return -1;

	add_buffer(sock);

	return 0;
}


/*
 * Reliably send 'msg' of length 'len' through previously connected
 * socket 'sock', subject to caveats contained in 'flags'. This function
 * will fragment arbitrarily large buffers. It returns 0 on success and
 * -1 on failure.
 */
int mt_send(int sock, void* msg, int len, int flags)
{
	int	mtu, tosend, bytesleft, sent, res;
	char	*d = msg;

	if (len == 0)
		return 0;

	mtu = rm_get_mtu(sock);
	bytesleft = len;
	sent = 0;

	/*
	 * Fragment large buffer into mtu size messages
	 */
	while (bytesleft > 0) {
		tosend = (bytesleft > mtu ? mtu : bytesleft);
		res = rm_send(sock, d+sent, tosend, flags);
		if (res < 0) return -1;
		sent += tosend;
		bytesleft -= tosend;
	}

	return len;
}


/*
 * Reliably send 'msg' of length 'len' through previously connected
 * socket 'sock'. This function will fragment arbitrarily large buffers. 
 * It returns 0 on success and -1 on failure.
 */
int mt_write(int sock, void* msg, int len)
{
	return mt_send(sock, msg, len, 0);
}


/*
 * Receive at most 'len' bytes into the buffer 'dest' from connected socket
 * 'sock', subject to options in 'flags'. Returns number of bytes received
 * or -1 on failure.
 */
int mt_recv(int sock, void* dest, int len, int flags)
{
	int		res;
	conn_buffer	*b;
	struct sockaddr_in from;

	b = get_buffer(sock);
	if (b == NULL) {
		errno = ENOTCONN;
		return -1;
	}

	if (b->bytes == 0) {
		/*
		 * We're not currently buffering any data. Get a fresh
		 * incoming packet. If len is bigger than the bytes 
		 * available in the incoming packet, copy the whole
		 * packet straight into the user buffer. Otherwise, copy
		 * it into our buffer and then a portion into user buffer.
		 */
		if (len >= rm_bytes_ready(sock)) {
			/*
			 * There's room enough for whole avail packet
			 */
			res = rm_recv(sock, dest, len, flags);
			return res;
		}
		else {
			/*
			 * Receive into our buffer, then copy to user buf.
			 * If the user is peeking the msg, we have to consume
			 * the message on the socket, and let them peek it
			 * from our local buffer.
			 */
			res = rm_recv(sock,b->buf,MT_MAX_MTU,flags&(~MSG_PEEK));
			if (res < 0) return -1;
			bcopy(b->buf, dest, len);
			if (flags & MSG_PEEK) {
				b->bytes = res;
				b->index = 0;
			}
			else {
				b->bytes = res - len;
				b->index = len;
			}
			return len;
		}
	}
	else if (b->bytes > len) {
		/*
		 * We already have more bytes available in our buffer
		 * than the user has requested. We can satisfy user's
		 * request immediately.
		 */
		bcopy(b->buf + b->index, dest, len);
		if (!(flags & MSG_PEEK)) {
			b->bytes -= len;
			b->index += len;
		}
		return len;
	}
	else {
		/*
		 * User wants more bytes than we have handy, but we'll
		 * just give 'em what we've got now.
		 */
		bcopy(b->buf + b->index, dest, b->bytes);
		res = b->bytes;
		if (!(flags & MSG_PEEK)) {
			b->bytes = 0;
			b->index = 0;
		}
		return res;
	}
}


/*
 * Receive at most 'len' bytes into the buffer 'dest' from connected socket
 * 'sock'. Returns number of bytes received or -1 on failure.
 */
int mt_read(int sock, void* b, int len)
{
	return mt_recv(sock, b, len, 0);
}


/*
 * Close a connected socket. Closing a connected socket involves initiating
 * a termination protocol that ensures that all sent messages are safely
 * delivered. Unfortunately, without a coordinator process, it's 
 * impossible to guarantee that both sides of this connection terminate
 * together when packets can be lost (it's the two-generals problem). 
 * However, the protocol we use is pretty good and will almost always work.
 * Only one side of the connection should initiate connection. The other
 * side should not call mt_close until it has received a zero length message
 * to indicate EOF.
 */
int mt_close(int sock)
{
	int	res;

	if (!rm_conn_closing(sock)) {  
		/* initiator of tear down protocol */

		/*
		 * Wait for all outgoing messages to be acked.
		 */
		rm_drain(sock);

		/*
		 * Send EOF message.
		 */
		rm_send(sock, "", 0, MSG_EOF);

		/*
		 * Wait for EOF to be acked.
		 */
		rm_drain(sock);
	}
	else {
		/* subordinate in tear down protocol */

		/*
		 * Wait for all outgoing messages to be acked.
		 */
		rm_drain(sock);

		/*
		 * Send ack back to initiator. We'll send it a few
		 * times to help guarantee delivery. This protocol is
		 * sensitive to the loss of this ack.
		 */
		rm_ack(sock);
		rm_ack(sock);
		rm_ack(sock);
	}

	delete_buffer(sock);
	return rm_close(sock);
}


